-------------------------------------------------------------------------------------
-- Project    : CevTes RVU (Runtime Verification Unit)
-------------------------------------------------------------------------------------
-- File       : ringbuffer.vhd
-- Author     : Thomas Reinbacher    
-- Copyright  : 2011, Thomas Reinbacher (treinbacher@ecs.tuwien.ac.at)
--              Vienna University of Technology, ECS Group
--              http://ti.tuwien.ac.at/ecs/research/projects/cevtes
-------------------------------------------------------------------------------------
-- Description: An event triggered ringbuffer datastructure to evaluate ptLTL# formulae
------------------------------------------------------------------------------------- 
library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.math_pkg.all;
use work.ram_with_double_memory_select_pkg.all;
use work.ringbuffer_pkg.all;
use work.cevRVU_top_pkg.all;
use work.testbench_util_pkg.all;

entity ringbuffer is
  
  generic (
    RAM_BLOCKS : integer
    );
  port (
    clk               : in  std_logic;
    reset_n           : in  std_logic;
    en                : in  std_logic;
    write             : in  std_logic;
    write_busy        : out std_logic;
    full              : out std_logic;
    obs_timestamp     : in  std_logic_vector(TIMESTAMP_WIDTH-1 downto 0);
    timestamp_start   : in  std_logic_vector(TIMESTAMP_WIDTH-1 downto 0);
    timestamp_end     : in  std_logic_vector(TIMESTAMP_WIDTH-1 downto 0);
    interval_min      : in  std_logic_vector(TIMESTAMP_WIDTH-1 downto 0);
    interval_max      : in  std_logic_vector(TIMESTAMP_WIDTH-1 downto 0);
    operator          : in  std_logic;
    check_validity_en : in  std_logic;
    gc_en             : in  std_logic;
    gc_fin            : out std_logic;
    valid             : out std_logic
    );

end ringbuffer;
architecture rtl of ringbuffer is

  type RB_WRITE_STATE is(STATE_WR_WAIT,
                         STATE_WR_IDLE,
                         STATE_WR_FETCH_DATA,
                         STATE_WR_CHECK_FEASIBLE,
                         STATE_WR_WRITE);
  signal wr_state, wr_state_next : RB_WRITE_STATE;

  type RB_GC_STATE is(STATE_GC_IDLE,
                      STATE_GC_WAIT,
                      STATE_GC_READ_LINE,
                      STATE_GC_EVAL_GARBAGE,
                      STATE_GC_FETCH_OBSERVATION_TIMESTAMP,
                      STATE_GC_EVAL_GARBAGE_COL_UNDER_RP,
                      STATE_GC_CHECK_PRE_VALID_SND_CONJ,
                      STATE_GC_SET_READ_PONITER,
                      STATE_GC_SET_READ_ADDRESS,
                      STATE_GC_CHECK_VALID,
                      STATE_GC_GOTO_NEXT_RING,
                      STATE_GC_ASSERT_FIN);
  signal gc_state, gc_state_next : RB_GC_STATE;

  type write_registers_type is record
    block_mem_select : std_logic_vector(1 downto 0);
    block_wr_address : std_logic_vector(log2c(TOTAL_ELEMENTS/RAM_BLOCKS) -1 downto 0);
    block_wr_data    : std_logic_vector(2*TIMESTAMP_WIDTH-1 downto 0);
    block_wr         : std_logic;
    wr_ptr           : ringbuffer_pointer;
    write_busy       : std_logic;
    feasible         : std_logic;
    virgin           : std_logic;  -- '1' when no element was written into ring buffer
  end record;
  signal s_wr_reg, s_wr_reg_next : write_registers_type;

  constant S_WRITE_REG_DEFAULT : write_registers_type :=
    (
      block_mem_select => (others => '0'),
      block_wr_address => (others => '0'),
      block_wr_data    => (others => '0'),
      block_wr         => '0',
      wr_ptr           => ("001", "000", "000"),
      write_busy       => '0',
      feasible         => '0',
      virgin           => '1'
      );       

  type gc_registers_type is record
    block_mem_select : std_logic_vector(1 downto 0);
    block_rd_address : std_logic_vector(log2c(TOTAL_ELEMENTS/RAM_BLOCKS) -1 downto 0);
    block_rd         : std_logic;
    rd_ptr           : ringbuffer_pointer;
    nxt_rd_ptr       : ringbuffer_pointer;
    cmp_result       : std_logic_vector(RAM_BLOCKS downto 0);  -- cmp(0) is a helper bit
    ring             : integer range 0 to 4;
    obs_timestamp    : std_logic_vector(TIMESTAMP_WIDTH-1 downto 0);
    timestamp        : ringbuffer_mem_line;
    snd_conj_valid   : std_logic;
    valid            : std_logic;
    gc_fin           : std_logic;
    helper           : std_logic_vector(NUM_OF_RINGS-1 downto 0);
    line_zero        : std_logic_vector(RAM_BLOCKS-1 downto 0);
  end record;
  signal s_gc_reg, s_gc_reg_next : gc_registers_type;

  signal current_timestamp_start : ringbuffer_half_mem_line := (others => (others => '0'));
  signal current_timestamp_end   : ringbuffer_half_mem_line := (others => (others => '0'));

  constant S_GC_REG_DEGAULT : gc_registers_type :=
    (
      block_mem_select => (others => '0'),
      block_rd_address => (others => '0'),
      block_rd         => '0',
      rd_ptr           => ("001", "000", "000"),
      nxt_rd_ptr       => (others => (others => '0')),
      cmp_result       => (others => '0'),
      ring             => 0,
      obs_timestamp    => (others => '0'),
      timestamp        => (others => (others => '0')),
      snd_conj_valid   => '0',
      valid            => '0',
      gc_fin           => '0',
      helper           => "101",
      line_zero        => (others => '0')
      );

  signal s_block_wr_address : std_logic_vector(log2c(TOTAL_ELEMENTS/RAM_BLOCKS) -1 downto 0);
  signal s_block_wr_data    : std_logic_vector(2*TIMESTAMP_WIDTH-1 downto 0);
  signal s_block_wr         : std_logic;

  signal s_block_rd_address : std_logic_vector(log2c(TOTAL_ELEMENTS/RAM_BLOCKS) -1 downto 0);
  signal s_block_rd_data    : block_data_mem;
  signal s_block_rd         : std_logic;

  signal s_wr_memory_select : std_logic_vector(log2c(RAM_BLOCKS)-1 downto 0);
  signal s_rd_memory_select : std_logic_vector(log2c(RAM_BLOCKS)-1 downto 0);

  -- for simulation only, comment this for synthesis
  signal s_free_elements : integer range 0 to TOTAL_ELEMENTS := 0;

  constant TIMESTAMP_LINE_ALL_ZERO : ringbuffer_entry := (others => '0');
begin  -- rtl

-- generate and connect ringbuffer ram RAM blocks
  gen_ram_blocks : for i in RAM_BLOCKS -1 downto 0 generate
    i_ram_with_double_memory_select : ram_with_double_memory_select
      generic map(
        ADDR_WIDTH             => log2c(TOTAL_ELEMENTS/RAM_BLOCKS),
        DATA_WIDTH             => 2*TIMESTAMP_WIDTH,
        MEMBUS_WR_SELECT_WIDTH => log2c(RAM_BLOCKS),
        MEMBUS_RD_SELECT_WIDTH => log2c(RAM_BLOCKS),
        MEMBUS_RD_SELECT_ID    => "11",
        MEMBUS_WR_SELECT_ID    => std_logic_vector(to_unsigned(i, log2c(RAM_BLOCKS)))
        )
      port map(
        clk              => clk,
        raddr            => s_block_rd_address,
        rdata            => s_block_rd_data(i),
        rd               => s_block_rd,
        rd_memory_select => s_rd_memory_select,
        waddr            => s_block_wr_address,
        wdata            => s_block_wr_data,
        wr               => s_block_wr,
        wr_memory_select => s_wr_memory_select
        );
  end generate;

  s_block_wr         <= s_wr_reg.block_wr;
  s_block_wr_address <= s_wr_reg.block_wr_address;
  s_block_wr_data    <= s_wr_reg.block_wr_data;
  s_wr_memory_select <= s_wr_reg.block_mem_select;
  s_rd_memory_select <= s_gc_reg.block_mem_select;
  s_block_rd         <= s_gc_reg.block_rd;
  s_block_rd_address <= s_gc_reg.block_rd_address;

  -- Full Detection
  p_full_detection : process(s_wr_reg.wr_ptr, s_gc_reg.rd_ptr, s_wr_reg.virgin)
  begin
    if s_wr_reg.virgin = '1' then
      full <= '0';
    else
      full <= compare_ptr(s_wr_reg.wr_ptr, s_gc_reg.rd_ptr);
    end if;
  end process p_full_detection;

  -- report free elements
  s_free_elements <= get_free_elements(s_wr_reg.wr_ptr, s_gc_reg.rd_ptr, s_wr_reg.virgin);

  ----------------------------------
  -- RINGBUFFER WRITE STATE MACHINE
  ----------------------------------
  write_next_state_logic : process(wr_state, en, write)
  begin
    wr_state_next <= wr_state;
    case wr_state is
      when STATE_WR_IDLE =>
        if en = '1' then
          wr_state_next <= STATE_WR_WAIT;
        end if;
        
      when STATE_WR_WAIT =>
        if write = '1' then
          wr_state_next <= STATE_WR_FETCH_DATA;
        end if;
        
      when STATE_WR_FETCH_DATA =>
        wr_state_next <= STATE_WR_CHECK_FEASIBLE;
        
      when STATE_WR_CHECK_FEASIBLE =>
        if(s_wr_reg.feasible = '1') then
          wr_state_next <= STATE_WR_WRITE;
        else
          wr_state_next <= STATE_WR_WAIT;
        end if;
        
      when STATE_WR_WRITE =>
        wr_state_next <= STATE_WR_WAIT;
      when others => null;
    end case;
  end process write_next_state_logic;

  -- output logic
  write_output : process(wr_state, s_wr_reg, timestamp_start, timestamp_end, s_wr_reg_next, interval_max, interval_min)
  begin
    s_wr_reg_next            <= s_wr_reg;
    s_wr_reg_next.block_wr   <= '0';
    s_wr_reg_next.write_busy <= '1';

    case wr_state is
      when STATE_WR_IDLE | STATE_WR_WAIT =>
        s_wr_reg_next.write_busy <= '0';
        null;
      when STATE_WR_FETCH_DATA =>
        s_wr_reg_next.block_wr_data <= timestamp_start & timestamp_end;
        -- feasible check
        if((unsigned(timestamp_end) - unsigned(timestamp_start)) >= (unsigned(interval_max) - unsigned(interval_min))) then
          s_wr_reg_next.feasible <= '1';
        else
          s_wr_reg_next.feasible <= '0';
        end if;

      when STATE_WR_CHECK_FEASIBLE =>
        null;

      when STATE_WR_WRITE =>
        s_wr_reg_next.block_wr_address <= slv(to_unsigned(get_row(s_wr_reg.wr_ptr), s_wr_reg_next.block_wr_address'length));
        s_wr_reg_next.block_mem_select <= slv(to_unsigned(get_col(s_wr_reg.wr_ptr), s_wr_reg_next.block_mem_select'length));
        s_wr_reg_next.block_wr         <= '1';
        s_wr_reg_next.wr_ptr           <= increment_ptr(s_wr_reg.wr_ptr);
        s_wr_reg_next.virgin           <= '0';
    end case;
  end process write_output;

  write_sync : process(clk, reset_n)
  begin
    if reset_n = '0' then
      s_wr_reg <= S_WRITE_REG_DEFAULT;
      wr_state <= STATE_WR_IDLE;
    elsif clk'event and clk = '1' then
      s_wr_reg <= s_wr_reg_next;
      wr_state <= wr_state_next;
    end if;
  end process write_sync;

  ----------------------------------------
  -- GARBAGE Collection State Machine
  ----------------------------------------
  gc_next_state_logic : process(gc_state, en, gc_en, s_gc_reg, check_validity_en)
  begin
    gc_state_next <= gc_state;
    case gc_state is

      when STATE_GC_IDLE =>
        if en = '1' then
          gc_state_next <= STATE_GC_WAIT;
        end if;

      when STATE_GC_WAIT =>
        if gc_en = '1' then
          gc_state_next <= STATE_GC_FETCH_OBSERVATION_TIMESTAMP;
        end if;

      when STATE_GC_FETCH_OBSERVATION_TIMESTAMP =>
        gc_state_next <= STATE_GC_READ_LINE;

      when STATE_GC_READ_LINE =>
        gc_state_next <= STATE_GC_EVAL_GARBAGE;

      when STATE_GC_EVAL_GARBAGE =>
        gc_state_next <= STATE_GC_EVAL_GARBAGE_COL_UNDER_RP;

      when STATE_GC_EVAL_GARBAGE_COL_UNDER_RP =>
        gc_state_next <= STATE_GC_SET_READ_ADDRESS;

      when STATE_GC_SET_READ_ADDRESS =>
        gc_state_next <= STATE_GC_GOTO_NEXT_RING;

      when STATE_GC_GOTO_NEXT_RING =>
        -- last ring
        if(s_gc_reg.ring = NUM_OF_RINGS-1) or
          -- cmp result was zero (all new elements) but there is no read pointer for this ring
          (s_gc_reg.cmp_result = "10000" and s_gc_reg.rd_ptr(s_gc_reg.ring+1) = "000") then
          gc_state_next <= STATE_GC_SET_READ_PONITER;
        else
          gc_state_next <= STATE_GC_READ_LINE;
        end if;
        
      when STATE_GC_SET_READ_PONITER =>
        gc_state_next <= STATE_GC_CHECK_PRE_VALID_SND_CONJ;

      when STATE_GC_CHECK_PRE_VALID_SND_CONJ =>
        gc_state_next <= STATE_GC_CHECK_VALID;

      when STATE_GC_CHECK_VALID =>
        gc_state_next <= STATE_GC_WAIT;

      when STATE_GC_ASSERT_FIN =>
        gc_state_next <= STATE_GC_WAIT;

    end case;
  end process gc_next_state_logic;

  gc_output_logic : process(gc_state, s_block_rd_data, s_gc_reg, obs_timestamp, interval_min,
                            current_timestamp_start, operator, interval_max, s_gc_reg_next, current_timestamp_end)
  begin
    s_gc_reg_next                  <= s_gc_reg;
    s_gc_reg_next.block_rd         <= '1';
    s_gc_reg_next.block_mem_select <= "11";
    s_gc_reg_next.valid            <= '0';
    s_gc_reg_next.gc_fin           <= '0';

    case gc_state is

      when STATE_GC_IDLE | STATE_GC_WAIT =>
        s_gc_reg_next.ring             <= 0;
        s_gc_reg_next.block_rd_address <= (others => '0');
        s_gc_reg_next.helper           <= "111";
        null;

      when STATE_GC_FETCH_OBSERVATION_TIMESTAMP =>
        s_gc_reg_next.obs_timestamp <= obs_timestamp;

      when STATE_GC_READ_LINE =>
        -- parallel read
        for j in 0 to 3 loop
          -- only read data if there is any
          if(s_block_rd_data(j) = TIMESTAMP_LINE_ALL_ZERO) then
            s_gc_reg_next.line_zero(j) <= '1';
          else
            s_gc_reg_next.timestamp(j) <= s_block_rd_data(j);
            current_timestamp_start(j) <= s_block_rd_data(j)(2*TIMESTAMP_WIDTH-1 downto TIMESTAMP_WIDTH);
            current_timestamp_end(j)   <= s_block_rd_data(j)(TIMESTAMP_WIDTH-1 downto 0);
            s_gc_reg_next.line_zero(j) <= '0';
          end if;
        end loop;
        
      when STATE_GC_EVAL_GARBAGE =>
        -- generate parallel comparators
        s_gc_reg_next.cmp_result(0) <= '1';
        for j in 0 to 3 loop
          -- Check: T.t_e < n - min
          if((unsigned(s_gc_reg.timestamp(j)(TIMESTAMP_WIDTH-1 downto 0)) < (unsigned(s_gc_reg.obs_timestamp) - unsigned(interval_min)))
             and
             (s_gc_reg_next.line_zero(j) = '0'))
          then
            s_gc_reg_next.cmp_result(1+j) <= '1';
          else
            s_gc_reg_next.cmp_result(1+j) <= '0';
          end if;
        end loop;

      when STATE_GC_EVAL_GARBAGE_COL_UNDER_RP =>
        -- The column under the Read Pointer is already too old, so we can find the oldest
        -- column and continue;
        if(s_gc_reg.cmp_result(get_pos_for_ring(s_gc_reg.rd_ptr, s_gc_reg.ring)) = '1') then
          -- find the rightmost '1' in cmp_result
          s_gc_reg_next.nxt_rd_ptr(s_gc_reg.ring) <= slv(to_unsigned(get_rightmost_one_pos(s_gc_reg.cmp_result(4 downto 1)), s_gc_reg_next.nxt_rd_ptr(0)'length));
          s_gc_reg_next.helper(s_gc_reg.ring)     <= '0';
          -- The column under the Read Pointer is not too old, we can possibly only refine
          -- this column a bit
        else
          s_gc_reg_next.nxt_rd_ptr(s_gc_reg.ring) <= slv(to_unsigned(get_pos_for_ring(s_gc_reg.rd_ptr, s_gc_reg.ring), s_gc_reg.nxt_rd_ptr(0)'length));
          s_gc_reg_next.helper(s_gc_reg.ring)     <= '1';
        end if;

      when STATE_GC_SET_READ_ADDRESS =>
        s_gc_reg_next.block_rd_address <= slv(to_unsigned(get_row_for_next_ring(s_gc_reg.nxt_rd_ptr, s_gc_reg.ring), s_gc_reg.block_rd_address'length));

      when STATE_GC_GOTO_NEXT_RING =>
        s_gc_reg_next.ring <= s_gc_reg.ring + 1;

      when STATE_GC_SET_READ_PONITER =>
        if(s_gc_reg.nxt_rd_ptr /= s_gc_reg.rd_ptr) then
          s_gc_reg_next.rd_ptr <= increment_ptr(s_gc_reg.nxt_rd_ptr);
        end if;
        if (s_gc_reg.helper = "000") then
          s_gc_reg_next.rd_ptr <= increment_ptr(s_gc_reg.rd_ptr);
        end if;

      when STATE_GC_CHECK_PRE_VALID_SND_CONJ =>
        if(unsigned(interval_min) > unsigned(s_gc_reg.obs_timestamp)) then
          s_gc_reg_next.snd_conj_valid <= '1';
        else
          if(unsigned(current_timestamp_end(get_pos_for_ring(s_gc_reg.rd_ptr, get_ring(s_gc_reg.rd_ptr))-1)) >=
             (unsigned(s_gc_reg.obs_timestamp) - unsigned(interval_min))) then
            s_gc_reg_next.snd_conj_valid <= '1';
          else
            s_gc_reg_next.snd_conj_valid <= '0';
          end if;
        end if;

      when STATE_GC_CHECK_VALID =>
        if(unsigned(s_gc_reg.obs_timestamp) > unsigned(interval_max)) then
          if(unsigned(current_timestamp_start(get_pos_for_ring(s_gc_reg.rd_ptr, get_ring(s_gc_reg.rd_ptr))-1)) <=
             (unsigned(s_gc_reg.obs_timestamp) - unsigned(interval_max))) then
            s_gc_reg_next.valid <= (s_gc_reg.snd_conj_valid and '1') xor operator;
          else
            s_gc_reg_next.valid <= '0' xor operator;
          end if;
        else
          if(current_timestamp_start(get_pos_for_ring(s_gc_reg.rd_ptr, get_ring(s_gc_reg.rd_ptr))-1) = C_TIMESTAMP_ALL_ZERO) then
            s_gc_reg_next.valid <= (s_gc_reg.snd_conj_valid and '1') xor operator;
          else
            s_gc_reg_next.valid <= '0' xor operator;
          end if;
        end if;
        s_gc_reg_next.gc_fin <= '1';

      when STATE_GC_ASSERT_FIN =>
        s_gc_reg_next.gc_fin <= '1';

    end case;
  end process gc_output_logic;

  gc_sync : process(clk, reset_n)
  begin
    if reset_n = '0' then
      s_gc_reg <= S_GC_REG_DEGAULT;
      gc_state <= STATE_GC_IDLE;
    elsif clk'event and clk = '1' then
      s_gc_reg <= s_gc_reg_next;
      gc_state <= gc_state_next;
    end if;
  end process gc_sync;

  valid      <= s_gc_reg.valid;
  gc_fin     <= s_gc_reg.gc_fin;
  write_busy <= s_wr_reg.write_busy;

end rtl;
