-------------------------------------------------------------------------------------
-- Project    : CevTes RVU (Runtime Verification Unit)
-------------------------------------------------------------------------------------
-- File       : ringbuffer_pkg.vhd
-- Author     : Thomas Reinbacher    
-- Copyright  : 2011, Thomas Reinbacher (treinbacher@ecs.tuwien.ac.at)
--              Vienna University of Technology, ECS Group
--              http://ti.tuwien.ac.at/ecs/research/projects/cevtes
-------------------------------------------------------------------------------------
-- Description:  Definitions and functions for the ringbuffer design
-------------------------------------------------------------------------------------    

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.math_pkg.all;
use work.CevRVU_top_pkg.all;

package ringbuffer_pkg is

  constant NUM_OF_BASE_ELEMENTS : natural   := 4;
  constant RAM_BLOCKS           : natural   := NUM_OF_BASE_ELEMENTS;
  constant NUM_OF_RINGS         : natural   := 3;
  constant OP_DIAMOND_DOT       : std_logic := '1';
  constant OP_BOX_DOT           : std_logic := '0';
  constant C_TIMESTAMP_ALL_ZERO   : std_logic_vector(TIMESTAMP_WIDTH-1 downto 0) := (others =>'0');
  constant TOTAL_ELEMENTS       : natural   := (NUM_OF_BASE_ELEMENTS)
                                               + 2** log2c(NUM_OF_BASE_ELEMENTS) * NUM_OF_BASE_ELEMENTS
                                               + 2** log2c(NUM_OF_BASE_ELEMENTS) * 4*NUM_OF_BASE_ELEMENTS;

  subtype ringbuffer_pointer_entry is std_logic_vector(2 downto 0);
  type    ringbuffer_pointer is array(0 to 2) of ringbuffer_pointer_entry;

  subtype ringbuffer_entry is std_logic_vector(2*TIMESTAMP_WIDTH -1 downto 0);
  subtype ringbuffer_half_entry is std_logic_vector(TIMESTAMP_WIDTH -1 downto 0);
  type    ringbuffer_mem_line is array(0 to 3) of ringbuffer_entry;
  type    ringbuffer_half_mem_line is array(0 to 3) of ringbuffer_half_entry;

  subtype pointer_entry is std_logic_vector(log2c(TOTAL_ELEMENTS/RAM_BLOCKS) -1 downto 0);
  type    rb_pointer_mem is array(0 to RAM_BLOCKS-1) of pointer_entry;

  type block_data_mem is array(0 to RAM_BLOCKS-1) of ringbuffer_entry;

  subtype block_ctrl_entry is std_logic;
  type    block_ctrl_mem is array(0 to RAM_BLOCKS-1) of block_ctrl_entry;

  subtype slv is std_logic_vector;

  function get_ring(pointer                            : ringbuffer_pointer) return integer;
  function get_row(pointer                             : ringbuffer_pointer) return integer;
  function get_row_for_next_ring(pointer               : ringbuffer_pointer; ring : integer) return integer;
  function get_col(pointer                             : ringbuffer_pointer) return integer;
  function get_pos_for_ring(pointer                    : ringbuffer_pointer; ring : integer) return integer;
  function increment_ptr(pointer                       : ringbuffer_pointer) return ringbuffer_pointer;
  function get_rightmost_one_pos(s                     : std_logic_vector) return integer;
  function compare_ptr(pointerA, pointerB              : ringbuffer_pointer) return std_logic;
  function get_free_elements(readPointer, writePointer : ringbuffer_pointer; virgin : std_logic) return integer;
end ringbuffer_pkg;

package body ringbuffer_pkg is

  function get_ring(pointer : ringbuffer_pointer) return integer is
    variable v_ring : natural                                        := 0;
    constant ZERO   : std_logic_vector(pointer(0)'length-1 downto 0) := (others => '0');
  begin
    if pointer(0) /= ZERO and pointer(1) /= ZERO and pointer(2) /= ZERO then
      v_ring := 2;
    elsif pointer(0) /= ZERO and pointer(1) /= ZERO and pointer(2) = ZERO then
      v_ring := 1;
    elsif pointer(0) /= ZERO and pointer(1) = ZERO and pointer(2) = ZERO then
      v_ring := 0;
    else
      assert false report"ring pointer invalid" severity error;
      return 0;
    end if;
    return v_ring;
  end function get_ring;

  function get_free_elements(readPointer, writePointer : ringbuffer_pointer; virgin : std_logic) return integer is
    variable v_freeElements : integer := 0;
    variable v_readPointer  : ringbuffer_pointer;
    variable v_writePointer : ringbuffer_pointer;
  begin
    v_readPointer  := readPointer;
    v_writePointer := writePointer;
    if(virgin = '1') then
      v_freeElements := TOTAL_ELEMENTS;
    elsif compare_ptr(readPointer, writePointer) = '1' then
      v_freeElements := 0;
    else
      for i in 1 to TOTAL_ELEMENTS loop
        v_writePointer := increment_ptr(v_writePointer);
        v_freeElements := v_freeElements + 1;
        if compare_ptr(v_readPointer, v_writePointer) = '1' then
          exit;
        end if;
      end loop;
    end if;
    return v_freeElements;
  end function;

  function get_row(pointer : ringbuffer_pointer) return integer is
    variable v_row  : integer := 0;
    variable v_ring : integer := 0;
  begin
    v_ring := get_ring(pointer);
    if v_ring = 0 then
      v_row := 0;
    elsif v_ring = 1 then
      v_row := to_integer(unsigned(pointer(0)));
    elsif v_ring = 2 then
      v_row := to_integer(unsigned(pointer(0))) * 4 + to_integer(unsigned(pointer(1)));
    end if;
    return v_row;
  end function get_row;

  function get_row_for_next_ring(pointer : ringbuffer_pointer; ring : integer) return integer is
    variable v_row : integer := 0;
  begin
    if ring = 0 then
      v_row := to_integer(unsigned(pointer(0)));
    elsif ring = 1 then
      --row := 8;
      v_row := to_integer(unsigned(pointer(0))) * 4 + to_integer(unsigned(pointer(1))) * 1;
    elsif ring = 2 then
      v_row := 0;
    end if;
    return v_row;
  end function get_row_for_next_ring;

  function get_col(pointer : ringbuffer_pointer) return integer is
    variable v_col : natural                                        := 0;
    constant ZERO  : std_logic_vector(pointer(0)'length-1 downto 0) := (others => '0');
  begin
    if pointer(0) /= ZERO and pointer(1) /= ZERO and pointer(2) /= ZERO then
      v_col := to_integer(unsigned(pointer(2))) - 1;
    elsif pointer(0) /= ZERO and pointer(1) /= ZERO and pointer(2) = ZERO then
      v_col := to_integer(unsigned(pointer(1))) - 1;
    elsif pointer(0) /= ZERO and pointer(1) = ZERO and pointer(2) = ZERO then
      v_col := to_integer(unsigned(pointer(0))) - 1;
    else
      v_col := 0;
    end if;
    return v_col;
  end function get_col;

  function get_pos_for_ring(pointer : ringbuffer_pointer; ring : integer) return integer is
    variable v_pos : integer := 0;
  begin
    if ring /= 0 and ring /= 1 and ring /= 2 then
      assert false report"given ring does not exist" severity error;
      return 0;
    end if;

    return to_integer(unsigned(pointer(ring)));
  end function get_pos_for_ring;

  function compare_ptr(pointerA, pointerB : ringbuffer_pointer) return std_logic is
    variable v_res : std_logic := '0';
  begin
    if pointerA(0) = pointerB(0) and pointerA(1) = pointerB(1) and pointerA(2) = pointerB(2) then
      v_res := '1';
    else
      v_res := '0';
    end if;
    return v_res;
  end function compare_ptr;

  function increment_ptr(pointer : ringbuffer_pointer) return ringbuffer_pointer is
    variable v_ptr : ringbuffer_pointer;
    constant ZERO  : std_logic_vector(v_ptr(0)'length-1 downto 0) := (others => '0');
  begin
    if pointer(2) = ZERO and pointer(1) = ZERO and pointer(0) = ZERO then
      assert false report"the ring buffer can never have the configuration 000 000 000 " severity error;
      return v_ptr;
    end if;

    if pointer(2) = ZERO and pointer(1) = ZERO then
      v_ptr(0) := pointer(0);
      v_ptr(1) := "001";
      v_ptr(2) := pointer(2);
      return v_ptr;
    end if;

    if pointer(2) /= "100" then
      v_ptr(0) := pointer(0);
      v_ptr(1) := pointer(1);
      v_ptr(2) := std_logic_vector(unsigned(pointer(2)) + 1);
      -- overflow
    else
      v_ptr(2) := ZERO;
      if pointer(1) /= "100" then
        v_ptr(0) := pointer(0);
        v_ptr(1) := std_logic_vector(unsigned(pointer(1)) + 1);
      else
        v_ptr(1) := ZERO;
        if pointer(0) /= "100" then
          v_ptr(0) := std_logic_vector(unsigned(pointer(0)) + 1);
        else
          -- wrap to first element in buffer
          v_ptr(0) := std_logic_vector(unsigned(ZERO) + 1);
        end if;
      end if;
    end if;
    return v_ptr;
  end function increment_ptr;

  function get_rightmost_one_pos(s : std_logic_vector) return integer is
    variable v_pos : integer := 0;
  begin
    for i in 1 to 4 loop
      if s(i) = '1' then
        v_pos := i;
      end if;
    end loop;
    return v_pos;
  end function get_rightmost_one_pos;

end package body ringbuffer_pkg;
