-------------------------------------------------------------------------------------
-- Project    : CevTes RVU (Runtime Verification Unit)
-------------------------------------------------------------------------------------
-- File       : ringbuffer.vhd
-- Author     : Thomas Reinbacher    
-- Copyright  : 2011, Thomas Reinbacher (treinbacher@ecs.tuwien.ac.at)
--              Vienna University of Technology, ECS Group
--              http://ti.tuwien.ac.at/ecs/research/projects/cevtes
-------------------------------------------------------------------------------------
-- Description: A time triggered ringbuffer datastructure to evaluate ptLTL# formulae
-------------------------------------------------------------------------------------   

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

use work.math_pkg.all;
use work.ram_pkg.all;
use work.ringbuffer_pkg.all;
use work.cevRVU_top_pkg.all;
use work.testbench_util_pkg.all;

entity ringbuffer is
  
  generic (
  RAM_BLOCKS : integer  -- (leave unconnected for time triggered ring buffer)
    );
  port (
    clk             : in  std_logic;
    reset_n         : in  std_logic;
    en              : in  std_logic;
    write           : in  std_logic;
    write_busy      : out std_logic;
    full            : out std_logic;
    rtc_timestamp   : in  std_logic_vector(TIMESTAMP_WIDTH-1 downto 0); 
    timestamp_start : in  std_logic_vector(TIMESTAMP_WIDTH-1 downto 0); -- T.\tau_s
    timestamp_end   : in  std_logic_vector(TIMESTAMP_WIDTH-1 downto 0); -- T.\tau_e
    interval_min    : in  std_logic_vector(TIMESTAMP_WIDTH-1 downto 0); -- min(J)
    interval_max    : in  std_logic_vector(TIMESTAMP_WIDTH-1 downto 0); -- max(J)
    operator        : in  std_logic;                                    
    gc_fin          : out std_logic;
    gc_en           : in  std_logic;
    gc_busy         : out std_logic;
    valid           : out std_logic
    );

end ringbuffer;
architecture rtl of ringbuffer is

  type RB_WRITE_STATE is(STATE_WR_WAIT,
                         STATE_WR_IDLE,
                         STATE_WR_FETCH_DATA,
                         STATE_WR_CHECK_FEASIBLE,
                         STATE_WR_WRITE);
  signal wr_state, wr_state_next : RB_WRITE_STATE;

  type RB_GC_STATE is(STATE_GC_IDLE,
                      STATE_GC_WAIT,
                      STATE_GC_WAIT_FOR_RAM,
                      STATE_GC_READ_TUPLE,
                      STATE_GC_EVAL_GARBAGE,
                      STATE_GC_BUFFER_EMPTY,
                      STATE_GC_CHECK_GARBAGE_AND_VALID
                      );
  signal gc_state, gc_state_next : RB_GC_STATE;

  type write_registers_type is record
    rb_wr_address : std_logic_vector(log2c(TOTAL_ELEMENTS) -1 downto 0);
    rb_wr_data    : std_logic_vector(2*TIMESTAMP_WIDTH-1 downto 0);
    rb_wr         : std_logic;
    wr_ptr        : ringbuffer_pointer;
    write_busy    : std_logic;
    feasible      : std_logic;
    force_gc      : std_logic;
    virgin        : std_logic;  
  end record;
  signal s_wr_reg, s_wr_reg_next : write_registers_type;

  constant S_WRITE_REG_DEFAULT : write_registers_type :=
    (
      rb_wr_address => (others => '0'),
      rb_wr_data    => (others => '0'),
      rb_wr         => '0',
      wr_ptr        => C_POINTER_ZERO,
      write_busy    => '0',
      feasible      => '0',
      force_gc      => '0',
      virgin        => '1'
      );       

  type gc_registers_type is record
    rb_rd_address   : std_logic_vector(log2c(TOTAL_ELEMENTS) -1 downto 0);
    rb_rd           : std_logic;
    rd_ptr          : ringbuffer_pointer;
    valid           : std_logic;
    gc_fin          : std_logic;
    gc_busy         : std_logic;
    snd_conj_valid  : std_logic;
    garbage         : std_logic;
    timestamp_start : std_logic_vector(TIMESTAMP_WIDTH-1 downto 0);
    timestamp_end   : std_logic_vector(TIMESTAMP_WIDTH-1 downto 0);
  end record;
  signal s_gc_reg, s_gc_reg_next : gc_registers_type;


  constant S_GC_REG_DEGAULT : gc_registers_type :=
    (
      rb_rd_address   => (others => '1'),
      rb_rd           => '0',
      rd_ptr          => C_POINTER_ZERO,
      valid           => '0',
      gc_fin          => '0',
      gc_busy         => '0',
      snd_conj_valid  => '0',
      garbage         => '0',
      timestamp_end   => (others => '0'),
      timestamp_start => (others => '0')
      );

  signal s_rb_wr_address : std_logic_vector(log2c(TOTAL_ELEMENTS) -1 downto 0);
  signal s_rb_wr_data    : std_logic_vector(2*TIMESTAMP_WIDTH-1 downto 0);
  signal s_rb_wr         : std_logic;

  signal s_rb_rd_address : std_logic_vector(log2c(TOTAL_ELEMENTS) -1 downto 0);
  signal s_rb_rd_data    : std_logic_vector(2*TIMESTAMP_WIDTH-1 downto 0);
  signal s_rb_rd         : std_logic;

  signal s_rtc_timestamp_last : std_logic_vector(TIMESTAMP_WIDTH-1 downto 0);
  signal s_rtc_change         : std_logic;

  -- for simulation only, comment this for synthesis
  signal s_free_elements : integer range 0 to TOTAL_ELEMENTS := 0;

  constant TIMESTAMP_LINE_ALL_ZERO : ringbuffer_entry := (others => '0');
begin  -- rtl

  -- connect ringbuffer memors (Block RAM)
  i_ringbuffer_ram : ram
    generic map(
      ADDR_WIDTH => log2c(TOTAL_ELEMENTS),
      DATA_WIDTH => 2* TIMESTAMP_WIDTH
      )
    port map(
      clk   => clk,
      raddr => s_rb_rd_address,
      rdata => s_rb_rd_data,
      rd    => s_rb_rd,
      waddr => s_rb_wr_address,
      wdata => s_rb_wr_data,
      wr    => s_rb_wr
      );

  s_rb_wr         <= s_wr_reg.rb_wr;
  s_rb_wr_address <= s_wr_reg.rb_wr_address;
  s_rb_wr_data    <= s_wr_reg.rb_wr_data;
  s_rb_rd         <= s_gc_reg.rb_rd;
  s_rb_rd_address <= s_gc_reg.rb_rd_address;

  -- Full Detection
  p_full_detection : process(s_wr_reg.wr_ptr, s_gc_reg.rd_ptr, s_wr_reg.virgin)
  begin
    if s_wr_reg.virgin = '1' then
      full <= '0';
    else
      full <= compare_ptr(s_wr_reg.wr_ptr, s_gc_reg.rd_ptr);
    end if;
  end process p_full_detection;

  -- Detect changes in rtc_timestamp, triggeres evaluation 
  process(clk, reset_n)
  begin
    if reset_n = '0' then -- active low reset
      s_rtc_timestamp_last <= (others => '0');
      s_rtc_change         <= '0';
    elsif clk'event and clk = '1' then
      if rtc_timestamp /= s_rtc_timestamp_last then
        s_rtc_change <= '1';
      else
        s_rtc_change <= '0';
      end if;
      s_rtc_timestamp_last <= rtc_timestamp;
    end if;
  end process;   
  
  -- report free elements
  s_free_elements <= get_free_elements(s_wr_reg.wr_ptr, s_gc_reg.rd_ptr, s_wr_reg.virgin);

  -----------------------------------
  -- RING BUFFER WRITE STATE MACHINE
  -----------------------------------
  write_next_state_logic : process(wr_state, en, write)
  begin
    wr_state_next <= wr_state;
    case wr_state is
      when STATE_WR_IDLE =>
        if en = '1' then
          wr_state_next <= STATE_WR_WAIT;
        end if;
        
      when STATE_WR_WAIT =>
        if write = '1' then
          wr_state_next <= STATE_WR_FETCH_DATA;
        end if;
        
      when STATE_WR_FETCH_DATA =>
        wr_state_next <= STATE_WR_CHECK_FEASIBLE;
        
      when STATE_WR_CHECK_FEASIBLE =>
        if(s_wr_reg.feasible = '1') then
          wr_state_next <= STATE_WR_WRITE;
        else
          wr_state_next <= STATE_WR_WAIT;
        end if;
        
      when STATE_WR_WRITE =>
        wr_state_next <= STATE_WR_WAIT;

      when others => null;
    end case;
  end process write_next_state_logic;

  -- output logic
  write_output : process(wr_state, s_wr_reg, timestamp_start, timestamp_end, s_wr_reg_next, interval_max, interval_min)
  begin
    s_wr_reg_next            <= s_wr_reg;
    s_wr_reg_next.rb_wr      <= '0';
    s_wr_reg_next.write_busy <= '1';
    s_wr_reg_next.force_gc   <= '0';

    case wr_state is
      when STATE_WR_IDLE | STATE_WR_WAIT =>
        s_wr_reg_next.write_busy <= '0';
        null;
      when STATE_WR_FETCH_DATA =>
        s_wr_reg_next.rb_wr_data <= timestamp_start & timestamp_end;
        -- feasible check
        if((unsigned(timestamp_end) - unsigned(timestamp_start)) >= (unsigned(interval_max) - unsigned(interval_min))) then
          s_wr_reg_next.feasible <= '1';
        else
          s_wr_reg_next.feasible <= '0';
        end if;

      when STATE_WR_CHECK_FEASIBLE =>
        null;

      when STATE_WR_WRITE =>
        s_wr_reg_next.rb_wr_address <= s_wr_reg.wr_ptr;
        s_wr_reg_next.rb_wr         <= '1';
        s_wr_reg_next.wr_ptr        <= increment_ptr(s_wr_reg.wr_ptr);
        s_wr_reg_next.virgin        <= '0';
        s_wr_reg_next.force_gc      <= '1';
    end case;
  end process write_output;

  -- sync state machine registers
  write_sync : process(clk, reset_n)
  begin
    if reset_n = '0' then
      s_wr_reg <= S_WRITE_REG_DEFAULT;
      wr_state <= STATE_WR_IDLE;
    elsif clk'event and clk = '1' then
      s_wr_reg <= s_wr_reg_next;
      wr_state <= wr_state_next;
    end if;
  end process write_sync;

  ----------------------------------------
  -- GARBAGE Collection State machine
  ----------------------------------------
  gc_next_state_logic : process(gc_state, en, s_gc_reg, s_wr_reg, s_rtc_change)
  begin
    gc_state_next <= gc_state;
    case gc_state is

      when STATE_GC_IDLE =>
        if en = '1' then
          gc_state_next <= STATE_GC_WAIT;
        end if;

      when STATE_GC_WAIT =>
        if s_rtc_change = '1' or s_wr_reg.force_gc = '1' then
          gc_state_next <= STATE_GC_WAIT_FOR_RAM;
        end if;

      when STATE_GC_WAIT_FOR_RAM =>
        gc_state_next <= STATE_GC_READ_TUPLE;

      when STATE_GC_READ_TUPLE =>
        if (compare_ptr(s_gc_reg.rd_ptr, s_wr_reg.wr_ptr) = '0') then
          gc_state_next <= STATE_GC_EVAL_GARBAGE;
        else
          gc_state_next <= STATE_GC_BUFFER_EMPTY;
        end if;

      when STATE_GC_EVAL_GARBAGE =>
        gc_state_next <= STATE_GC_CHECK_GARBAGE_AND_VALID;
        
      when STATE_GC_BUFFER_EMPTY =>
        gc_state_next <= STATE_GC_WAIT;

      when STATE_GC_CHECK_GARBAGE_AND_VALID =>
        gc_state_next <= STATE_GC_WAIT;

    end case;
  end process gc_next_state_logic;

  gc_output_logic : process(gc_state, s_rb_rd_data, s_gc_reg, interval_min,
                            operator, interval_max, s_gc_reg_next, s_wr_reg)
  begin
    -- defaults
    s_gc_reg_next         <= s_gc_reg;
    s_gc_reg_next.rb_rd   <= '1';
    s_gc_reg_next.gc_fin  <= '0';
    s_gc_reg_next.garbage <= '0';
    s_gc_reg_next.gc_busy <= '0';

    case gc_state is

      when STATE_GC_IDLE | STATE_GC_WAIT =>

      when STATE_GC_WAIT_FOR_RAM =>
        s_gc_reg_next.gc_busy <= '1';
        
      when STATE_GC_READ_TUPLE =>
        s_gc_reg_next.timestamp_start <= s_rb_rd_data(2*TIMESTAMP_WIDTH -1 downto TIMESTAMP_WIDTH);
        s_gc_reg_next.timestamp_end   <= s_rb_rd_data(TIMESTAMP_WIDTH-1 downto 0);
        s_gc_reg_next.gc_busy         <= '1';

      when STATE_GC_EVAL_GARBAGE =>
        -- calculate the second conjuct of the valid boxdot predicate
        if(unsigned(interval_min) > unsigned(rtc_timestamp))then
          s_gc_reg_next.snd_conj_valid <= '1';
          s_gc_reg_next.garbage        <= '0';
        else
          if(unsigned(s_gc_reg.timestamp_end) >= (unsigned(rtc_timestamp) - unsigned(interval_min))) then
            s_gc_reg_next.snd_conj_valid <= '1';
            s_gc_reg_next.garbage        <= '0';
          else
            s_gc_reg_next.snd_conj_valid <= '0';
            s_gc_reg_next.garbage        <= '1';
          end if;
        end if;
        s_gc_reg_next.rb_rd   <= '0';
        s_gc_reg_next.gc_busy <= '1';

      when STATE_GC_CHECK_GARBAGE_AND_VALID =>
        -- implementation of the garbage predicate
        -- the garbage predicate is the negation of the second conjunct of the valid boxdot predicate
        if(s_gc_reg.garbage = '1') then
          -- increment read pointer
          s_gc_reg_next.rd_ptr <= increment_read_ptr(s_gc_reg.rd_ptr, s_wr_reg.wr_ptr);
        end if;

        -- implementation of the valid predicate
        if(unsigned(rtc_timestamp) > unsigned(interval_max))then
          if(unsigned(s_gc_reg.timestamp_start) <= (unsigned(rtc_timestamp) - unsigned(interval_max))) then
            s_gc_reg_next.valid <= (s_gc_reg.snd_conj_valid and '1') xor operator;
          else
            s_gc_reg_next.valid <= '0' xor operator;
          end if;
        else
          if(s_gc_reg.timestamp_start = C_TIMESTAMP_ALL_ZERO) then
            s_gc_reg_next.valid <= (s_gc_reg.snd_conj_valid and '1') xor operator;
          else
            s_gc_reg_next.valid <= '0' xor operator;
          end if;
        end if;
        s_gc_reg_next.gc_fin        <= '1';
        s_gc_reg_next.rb_rd_address <= s_gc_reg.rd_ptr;
        s_gc_reg_next.gc_busy       <= '1';
        
      when STATE_GC_BUFFER_EMPTY =>
        s_gc_reg_next.gc_fin        <= '1';
        s_gc_reg_next.valid         <= '0';
        s_gc_reg_next.rb_rd_address <= s_gc_reg.rd_ptr;
        s_gc_reg_next.gc_busy       <= '1';
    end case;
  end process gc_output_logic;

  -- sync state machine registers
  gc_sync : process(clk, reset_n)
  begin
    if reset_n = '0' then
      s_gc_reg <= S_GC_REG_DEGAULT;
      gc_state <= STATE_GC_IDLE;
    elsif clk'event and clk = '1' then
      s_gc_reg <= s_gc_reg_next;
      gc_state <= gc_state_next;
    end if;
  end process gc_sync;

  valid      <= s_gc_reg.valid;
  gc_fin     <= s_gc_reg.gc_fin;
  write_busy <= s_wr_reg.write_busy;
  gc_busy    <= s_gc_reg.gc_busy;

end rtl;
