-------------------------------------------------------------------------------------
-- Project    : CevTes RVU (Runtime Verification Unit)
-------------------------------------------------------------------------------------
-- File       : ringbuffer_pkg.vhd
-- Author     : Thomas Reinbacher    
-- Copyright  : 2011, Thomas Reinbacher (treinbacher@ecs.tuwien.ac.at)
--              Vienna University of Technology, ECS Group
--              http://ti.tuwien.ac.at/ecs/research/projects/cevtes
-------------------------------------------------------------------------------------
-- Description: Definitions and functions for the ringbuffer design
------------------------------------------------------------------------------------- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use work.math_pkg.all;
use work.cevRVU_top_pkg.all;

package ringbuffer_pkg is

  constant OP_DIAMOND_DOT       : std_logic                                    := '1';
  constant OP_BOX_DOT           : std_logic                                    := '0';
  constant TOTAL_ELEMENTS       : natural                                      := 32;
  constant C_TIMESTAMP_ALL_ZERO : std_logic_vector(TIMESTAMP_WIDTH-1 downto 0) := (others => '0');

  subtype ringbuffer_pointer is std_logic_vector(log2c(TOTAL_ELEMENTS)-1 downto 0);
  constant C_POINTER_ZERO       : ringbuffer_pointer                           := (others => '0');
  
  subtype ringbuffer_entry is std_logic_vector(2*TIMESTAMP_WIDTH -1 downto 0);
  subtype ringbuffer_half_entry is std_logic_vector(TIMESTAMP_WIDTH -1 downto 0);
  subtype slv is std_logic_vector;

  function increment_ptr(pointer                       : ringbuffer_pointer) return ringbuffer_pointer;
  function increment_read_ptr(rd_pointer, wr_pointer   : ringbuffer_pointer) return ringbuffer_pointer;
  function compare_ptr(pointerA, pointerB              : ringbuffer_pointer) return std_logic;
  function get_free_elements(readPointer, writePointer : ringbuffer_pointer; virgin : std_logic) return integer;
  function is_full(rd_pointer, wr_pointer              : ringbuffer_pointer) return std_logic;


end ringbuffer_pkg;

package body ringbuffer_pkg is

  function get_free_elements(readPointer, writePointer : ringbuffer_pointer; virgin : std_logic) return integer is
    variable v_freeElements : integer := 0;
    variable v_readPointer  : ringbuffer_pointer;
    variable v_writePointer : ringbuffer_pointer;
  begin
    v_readPointer  := readPointer;
    v_writePointer := writePointer;
    if(virgin = '1') then
      v_freeElements := TOTAL_ELEMENTS;
    elsif compare_ptr(readPointer, writePointer) = '1' then
      v_freeElements := 0;
    else
      for i in 1 to TOTAL_ELEMENTS loop
        v_writePointer := increment_ptr(v_writePointer);
        v_freeElements := v_freeElements + 1;
        if compare_ptr(v_readPointer, v_writePointer) = '1' then
          exit;
        end if;
      end loop;
    end if;
    return v_freeElements;
  end function;

  function compare_ptr(pointerA, pointerB : ringbuffer_pointer) return std_logic is
    variable v_res : std_logic := '0';
  begin
    if pointerA = pointerB then
      v_res := '1';
    else
      v_res := '0';
    end if;
    return v_res;
  end function compare_ptr;

  function is_full(rd_pointer, wr_pointer : ringbuffer_pointer) return std_logic is
    variable v_is_full        : std_logic := '0';
    variable v_incremented_wr : ringbuffer_pointer;
  begin
    v_incremented_wr := increment_ptr(wr_pointer);
    if(compare_ptr(v_incremented_wr, rd_pointer) = '1')then
      v_is_full := '1';
    else
      v_is_full := '0';
    end if;
    return v_is_full;
  end function is_full;

  function increment_ptr(pointer : ringbuffer_pointer) return ringbuffer_pointer is
    variable v_ptr : ringbuffer_pointer;
  begin
    v_ptr := std_logic_vector(unsigned(pointer) + 1);
    return v_ptr;
  end function increment_ptr;

  function increment_read_ptr(rd_pointer, wr_pointer : ringbuffer_pointer) return ringbuffer_pointer is
    variable v_ptr : ringbuffer_pointer;
  begin
    if(compare_ptr(rd_pointer, wr_pointer) = '1')then
      v_ptr := rd_pointer;
    else
      v_ptr := increment_ptr(rd_pointer);
    end if;
    return v_ptr;
  end function increment_read_ptr;
end package body ringbuffer_pkg;
